#!/usr/bin/env python
"""
Bike IM TCP Protocol
by Javantea
April 12, 2011

Jabber takes too many libraries, IRC doesn't quite do what I need, SMTP 
requires a server and a client. So what do I do? I write yet another bad 
protocol. But don't worry, I'm building it based on interchangability.

1) Ease of SMTP (Connect, from, to, data, send disconnect)
2) Immediacy of IM (when the message is sent, it is available to the user.
3) Requires no new client like a boss.
4) Remembers when you disconnect to not lose all history.

How do I do it? Implement SMTP, HTTP, and a simple IM protocol in Python. Make 
sure every single line is perfect, then boom. By the way, this has to 
interoperate with KMFDM.

How does one person talk to another person?
HTTP or SMTP to the server.
The server then sees the person online and sends it along.

How does the server send multiple message?
IM client connects to server.
Server says: you have messages.
IM client retrieves them with a request.

Note: I could do xmpp because there is a reasonable xmppd written by xmpppy.
I won't though. =)

So what IM client should I use? Aw crap, SSH? IRC? Netcat seems the easiest for 
some reason... Well shit. IRC it is then =)

PING :843E553C
PONG :843E553C
:irc.mzima.net 433 * Javantea :Nickname is already in use.
:irc.mzima.net 375 Javantea_ :- irc.mzima.net Message of the Day - 
:irc.mzima.net 372 Javantea_ : hi there
:irc.mzima.net 376 Javantea_ :End of /MOTD command.
PING 1302672521
PONG 1302672521
JOIN #neg9
:Javantea_!~jvoss@suzy.altsci.com JOIN :#altsci
:irc.mzima.net 332 Javantea_ #altsci :AltSci is up, data is restored, user accounts are enabled.
:irc.mzima.net 333 Javantea_ #altsci Javantea!~r3wt@c-66-235-61-3.sea.wa.customer.broadstripe.net 1302504621
:irc.mzima.net 353 Javantea_ = #altsci :Javantea_ @Javantea
:irc.mzima.net 366 Javantea_ #altsci :End of /NAMES list.
PRIVMSG #altsci hi
:Javantea!~r3wt@c-66-235-61-3.sea.wa.customer.broadstripe.net PRIVMSG #altsci :hello

Wow that's easy...

JOIN #neg9
:Javantea!~jvoss@localhost JOIN :#neg9
:localhost 332 Javantea #neg9 :Hi there
:localhost 333 Javantea #neg9 Javantea!~r3wt@c-66-235-61-3.sea.wa.customer.broadstripe.net 1302504621
:localhost 353 Javantea= #neg9 :Javantea @b
:localhost 366 Javantea #neg9 :a

Not quite working.

:localhost 332 Javantea #neg9 :AltSci is up, data is restored, user accounts are enabled.
:localhost 333 Javantea #neg9 Javantea!~r3wt@c-66-235-61-3.sea.wa.customer.broadstripe.net 1302504621
:localhost 353 Javantea = #neg9 :Javantea @Javantea_
:localhost 366 Javantea #neg9 :End of /NAMES list.

:localhost 372 Javantea :-  768 days, 22 hours, 15 minutes and 24 seconds
:localhost 376 Javantea :End of /MOTD command.

:Javanteax!~r3wt@c-66-235-61-3.sea.wa.customer.broadstripe.net PRIVMSG #neg9 :hello and welcome

:Javantea_!~jvoss@suzy.altsci.com JOIN :#neg9
:irc.mzima.net 332 Javantea_ #neg9 :AltSci is up, data is restored, user accounts are enabled.
:irc.mzima.net 333 Javantea_ #neg9 Javantea!~r3wt@c-66-235-61-3.sea.wa.customer.broadstripe.net 1302504621
:irc.mzima.net 353 Javantea_ = #neg9 :Javantea_ @Javantea
:irc.mzima.net 366 Javantea_ #neg9 :End of /NAMES list.

Ready to rock.

State 1: Connected
Require NICK
Send PING
Accept PONG
Send $motd
Optional Store USER
Goto State 2

State 2:Valid Nick
Accept JOIN -> State 3
Accept PRIVMSG -> State 4
Accept PING -> Send PONG

State 3:In a chat room
Add room to user's subscriptions
Accept JOIN -> State 3
Accept PRIVMSG -> State 4
Accept PING -> Send PONG
Accept PART -> Send :javantea_!~jvoss@suzy.altsci.com PART #altsci

State 4:Sending A Msg
Send message to everyone in room identified by PRIVMSG
Goto State 3

How does the HTTP Server work? It's an IRC bot!
The cool thing about the HTTP server is that it fields requests, sending all 
along to the smarts.

This is going to be the dopest chat server ever.

Just need to make this super fast, ie right now.

So the IRC bot takes messages and sends them to the channel.
It also receives private messages and channel messages.
The commands it allows are:
messages aka /messages/
messages $n aka /messages/$n
m
m $n

TODO:
Authentication?
DoS Protection
IRC Bot is not stable enough
message delete
m | m $i | md $i
Close connections on exit
"""

# HTTP Client and supporting utility functions
import urllib

# HTTP Server
import BaseHTTPServer

# TCP Socket Server for IRC Server
import SocketServer
import socket

# Multi-threading for multiple servers
import threading
import time

# IRC Bot Client
import ircbot
from irclib import nm_to_n, nm_to_h, irc_lower

# reverse of the repr function, written by me, Javantea
import reverserepr

# This is the bot and is global so that we can use it.
ircb = None

class BikeIM_HTTPServer(BaseHTTPServer.HTTPServer):
	def __init__(self, address):
		BaseHTTPServer.HTTPServer.__init__(self, address, BikeIM_HTTPHandler)
	#end def __init__(self, address)
	
#end class BikeIM_HTTPServer(BaseHTTPServer.HTTPServer)

class BikeIM_HTTPHandler(BaseHTTPServer.BaseHTTPRequestHandler):
	
	def do_GET(self):
		global messages
		print 'http req', self.path, self.headers
		if self.path == '/':
			# Static index for now.
			self.send_response(200)
			self.end_headers()
			self.wfile.write(file('index.html', 'rb').read())
			return
		elif self.path[:10] == '/messages/':
			# Handle all /messages/
			if len(self.path) == 10:
				# /messages/
				self.send_response(200)
				self.end_headers()
				list_data = '<ul>'
				for message in messages:
					sub = message.subject
					if sub == '':
						sub = 'Untitled'
					#end if
					list_data += '<li><a href="/messages/' + str(message.id) + '">' + sub + '</a><form method="post" action="/messages/' + str(message.id) + '/delete" onsubmit="return confirmDelete();"><input type="submit" value="Delete" /></form></li>\n'
				#next message
				if list_data == '<ul>':
					list_data = 'No Messages Yet, come back soon.'
				else:
					list_data += '</ul>\n'
				#end if
				msgs = file('messages.html', 'rb').read()
				msgs = msgs.replace('<div id="list"></div>', '<div id="list">' + list_data + '</div>')
				self.wfile.write(msgs)
				return
			elif self.path[10:] == 'post':
				# /messages/post
				self.send_response(200)
				self.end_headers()
				self.wfile.write(file('post.html', 'rb').read())
				return
			#end if
			# /messages/$i
			try:
				msg_i = int(self.path[10:])
				msg_data = None
				for message in messages:
					if message.id != msg_i: continue
					msg_data = message
					break
				#next message
			except ValueError:
				msg_data = None
			#end try
			if msg_data == None:
				# /messages/bad
				self.send_response(404, 'not found')
				self.end_headers()
				self.wfile.write('Message not found')
				return
			#end if
			self.send_response(200)
			self.end_headers()
			msg = file('message.html', 'rb').read()
			msg = msg.replace('<div id="subject"></div>', '<div id="subject">' + sanitize(msg_data.subject) + '</div>')
			msg = msg.replace('<div id="type"></div>', '<div id="type">' + sanitize(msg_data.msgType) + '</div>')
			msg = msg.replace('<div id="date"></div>', '<div id="date">' + sanitize(time.ctime(msg_data.msgTime)) + '</div>')
			msg = msg.replace('<div id="msg"></div>', '<div id="msg">' + sanitize(msg_data.body) + '</div>')
			msg = msg.replace('<div id="owner"></div>', '<div id="owner">' + sanitize(msg_data.owner) + '</div>')
			self.wfile.write(msg)
			return
		else:
			# TODO: css files, etc.
			self.send_response(404, 'not found')
			self.end_headers()
			self.wfile.write('File not found')
			return
		#end if
	#end def do_GET()
	
	def do_POST(self):
		# TODO: Limit flood or large messages.
		global messages
		print 'http post', self.path, self.headers
		if self.path == '/messages/post':
			if 'Content-Length' not in self.headers:
				self.send_response(400, 'bad request')
				self.end_headers()
				self.wfile.write('No Data, missing Content-Length')
				return
			#end if
			try:
				data_len = int(self.headers['Content-Length'])
			except ValueError:
				data_len = None
			#end try
			if data_len == None or data_len > 100000:
				self.send_response(400, 'bad request')
				self.end_headers()
				self.wfile.write('Bad Content-Length')
				return
			#end if
			messages.append(BikeIM_Message(self.rfile.read(data_len)))
			if ircb:
				ircb.reply(ircb.message(str(messages[-1].id)), ircb.channel)
			else:
				print 'Warning: Missing irc bot.'
			#end if
			saveMessages(messages)
			self.send_response(200)
			self.end_headers()
			msg_data = "Success"
			msg = file('posted.html', 'rb').read()
			msg = msg.replace('<div id="msg"></div>', '<div id="msg">' + msg_data + '</div>')
			self.wfile.write(msg)
			return
		elif self.path[:10] == '/messages/':
			if len(self.path) > 17 and self.path[-7:] == '/delete':
				try:
					msg_id = int(self.path[10:-7])
				except ValueError:
					self.send_response(404, 'not found')
					self.end_headers()
					self.wfile.write('File not found')
					return
				#end try
				msg_ids = [message.id for message in messages]
				if msg_id not in msg_ids:
					self.send_response(404, 'not found')
					self.end_headers()
					self.wfile.write('File not found')
					return
				#end if
				messages.pop(msg_ids.index(msg_id))
				saveMessages(messages)
				self.send_response(200)
				self.end_headers()
				msg_data = 'Deleted %i successfully' % msg_id
				msg = file('deleted.html', 'rb').read()
				msg = msg.replace('<div id="msg"></div>', '<div id="msg">' + msg_data + '</div>')
				self.wfile.write(msg)
				return
			#end if
		else:
			self.send_response(404, 'not found')
			self.end_headers()
			self.wfile.write('File not found')
			return
		#end if
	#end def do_POST()
#end class BikeIM_HTTPHandler(BaseHTTPServer.BaseHTTPRequestHandler)

def saveMessages(messages):
	"""
	Saves message list to messages.txt in a text format that is easy to 
	read, edit, and load.
	"""
	f = file('messages.txt','w')
	f.write('Messages\n')
	for message in messages:
		f.write("%d %d %s %s %s %s\n" % (message.id, message.msgTime, repr(message.msgType), repr(message.subject), repr(message.body), repr(message.owner)))
	#next message
	f.write('End\n')
	f.close()
	
	# Save a copy that appends and never allows deletes.
	# This will get big over time, but sort messages_all.txt | uniq helps.
	# It puts it in order of id which is nice.
	f = file('messages_all.txt','a')
	f.write('Messages\n')
	for message in messages:
		f.write("%d %d %s %s %s %s\n" % (message.id, message.msgTime, repr(message.msgType), repr(message.subject), repr(message.body), repr(message.owner)))
	#next message
	f.write('End\n')
	f.close()
#end def saveMessages(messages)

def loadMessages():
	"""
	Loads message list from messages.txt using a text format that is easy to 
	read, edit, and load.
	"""
	fdata = file('messages.txt','r').read().split('\n')
	if len(fdata) <= 3:
		print "Error: Not enough lines to complete.", len(fdata)
		return []
	#end if
	header = fdata.pop(0)
	if header != 'Messages':
		print "Error: Header not Messages.", header
		return []
	#end if
	fdata.pop()
	footer = fdata.pop()
	if footer != 'End':
		print "Error: Footer not End.", footer
		return []
	#end if
	messages = []
	for msg in fdata:
		messages.append(BikeIM_Message(msg, 'file'))
	#next msg
	return messages
#end def loadMessages()

def sanitize(data):
	output = data.replace('>','&gt;').replace('<','&lt;').replace('&','&amp;')
	return output
#end def sanitize(data)

class BikeIM_Message:
	def __init__(self, inputData, inputType='url'):
		global currentMaxMsgId
		useId = currentMaxMsgId
		owner = ''
		dataDict = {}
		if inputType == 'url':
			data = map(urllib.unquote, inputData.split('&'))
			for item in data:
				if '=' not in item: continue
				key, value = item.split('=')
				dataDict[key.replace('+', ' ')] = value.replace('+', ' ')
			#next item
			dataDict['time'] = time.time()
		else:
			# inputType == file
			dataDict = readMessageRepr(inputData)
			# use id instead of currentMaxMsgId
			if 'id' in dataDict:
				try:
					useId = int(dataDict['id'])
				except ValueError:
					print 'Error: Invalid message id', dataDict['id']
				#end try
			#end if
			if 'owner' in dataDict:
				owner = dataDict['owner']
			#end if
		#end if
		subject = 'No Subject'
		msgType = ''
		body    = ''
		msgTime = 0
		if 'subject' in dataDict:
			subject = dataDict['subject']
		#end if
		if 'body' in dataDict:
			body = dataDict['body']
		#end if
		if 'type' in dataDict:
			msgType = dataDict['type']
		#end if
		if 'time' in dataDict:
			msgTime = dataDict['time']
		else:
			msgTime = time.time()
		#end if
		self.subject = subject
		self.body = body
		self.msgType = msgType
		self.msgTime = msgTime
		self.owner = owner
		# use id from file instead of currentMaxMsgId
		self.id = useId
		# Should we not increment? Should we not?
		currentMaxMsgId = max(currentMaxMsgId, useId + 1)
	#end def __init__(inputData)
#end class BikeIM_Message

def readMessageRepr(data):
	"""
	Reverse one integer and 3 repr
	"""
	if ' ' not in data: return {}
	data_list = data.split(' ', 2)
	if len(data_list) < 3: return
	msg_id, msg_time, data2 = data_list
	try:
		msg_time = int(msg_time)
	except:
		print 'Error: Invalid time', msg_time
		msg_time = 0
	#end try
	message_list = reverserepr.reverseRepr(data2, 'array')
	if len(message_list) == 7:
		return {'id':msg_id, 'time':msg_time, 'type':message_list[0], 'subject':message_list[2], 'body':message_list[4], 'owner':message_list[6]}
	elif len(message_list) == 5:
		return {'id':msg_id, 'time':msg_time, 'type':message_list[0], 'subject':message_list[1], 'body':message_list[2], 'owner':message_list[3]}
	#end if
	return {'id':msg_id, 'time':msg_time}
#end def readMessageRepr(data)

class BikeIM_IRCServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):
	def __init__(self, address):
		SocketServer.TCPServer.__init__(self, address, BikeIM_IRCHandler)
		self.serverName = 'test'
		self.motd_a = ['Greetings, welcome to this humble IRC server.',
		'Contact jvoss@altsci.com for information.']
		self.irc_clients = []
		self.channels = {}
		self.requests = []
	#end def __init__(self, address)
#end class BikeIM_IRCServer(SocketServer.TCPServer)

"""
	It seems that I don't need to override this because we get access to the server 
	via the self.server. Duh.
	def finish_request(self, request, client_address):
		"Finish one request by instantiating RequestHandlerClass.
		Copied directly from Python 2.6.6-r2 SocketServer.py.
		I am overriding this which is quite normal.
		"
		self.RequestHandlerClass(request, client_address, self)

"""

class BikeIM_IRCHandler(SocketServer.BaseRequestHandler):
	motd = [":%s 375 %s :- %s Message of the Day - ",
		":%s 372 %s : %s",
		":%s 376 %s :End of /MOTD command."]
	welcomeMsg = "Hello, welcome to BikeIM."
	def handle(self):
		"""
		The IRC Server spawns one handler per connection.
		TODO: 
		limit the amount of data to MB per minute.
		Use a config file instead of the below.
		"""
		data = ''
		state = 1
		self.username = 'unknown'
		username = self.username
		self.user_data = None
		self.host = socket.getnameinfo(self.client_address, socket.AF_INET)
		commandsSent = 0
		# Ugh
		self.exiting = False
		# Since we break, we don't need to do a check here.
		while True:
			try:
				data = self.request.recv(1024)
				if data == '':
					# They have disconnected.
					break
				#end if
				data = data.strip()
				data = data
			except Exception, e:
				# TODO: Figure out every possible exception.
				print e
				break
			#end if
			commandsSent += 1
			print "%s:%d wrote:" % self.client_address, data
			lines = [d.strip('\r') for d in data.split('\n')]
			cmds = [d[:4] for d in lines]
			# QUIT always exits no matter what we were supposed to do.
			if 'QUIT' in cmds:
				self.request.send(':%s!~unknown@%s QUIT :Client Quit\n' % (username, self.host[0]))
				break
			#end if
			if state == 1:
				# In state 1, we only allow them 10 commands
				if commandsSent > 10:
					self.request.send('ERROR :%s only allows 10 commands before a valid nick\n' % self.server.serverName)
					break
				#end if
				# Connected
				if 'USER' in cmds:
					# TODO: Parse user, save to logs
					# Format: USER jvoss ASLinWS01 localhost :purple
					self.user_data = lines[cmds.index('USER')].split(' ')
					self.user_data.pop(0)
				#end if
				# Make sure they sent the NICK command.
				if 'NICK' not in cmds:
					continue
				#end if
				# NICK username
				username = lines[cmds.index('NICK')][5:]
				# Ensure that the username isn't already taken.
				if username in self.server.irc_clients:
					self.request.send(':%s 433 * %s :Nickname is already in use.\n' % (self.server.serverName, username))
					continue
				#end if
				self.server.irc_clients.append(username)
				# Changed the username.
				self.username = username
				# Welcome is required for bots.
				self.request.send(':%s 001 %s :%s\n' % (self.server.serverName, username, self.welcomeMsg))
				# Send $motd
				motd_out = [self.motd[0] % (self.server.serverName, username, self.server.serverName)]
				for motd_line in self.server.motd_a:
					motd_out.append(self.motd[1] % (self.server.serverName, username, motd_line))
				#next motd_line
				motd_out.append(self.motd[2] % (self.server.serverName, username))
				# Every command ends with a newline
				self.request.send('\n'.join(motd_out) + '\n')
				# We can now add this to server.requests because it's valid.
				self.server.requests.append(self)
				# TODO: Sending a ping 
				# PONG can be parsed later.
				# Goto State 2: Valid Nick
				state = 2
			elif state == 2:
				# Valid Nick
				# In a chat room
				# Sending A Msg
				noCmd = True
				if 'JOIN' in cmds:
					state = 2
					self.joinChannel(username, lines[cmds.index('JOIN')])
					noCmd = False
				#end if
				if 'PRIV' in cmds:
					state = 2
					self.sendMsg(username, lines[cmds.index('PRIV')])
					noCmd = False
				#end if
				if 'ISON' in cmds:
					users = lines[cmds.index('ISON')].split(' ')
					users.pop(0)
					# Is a user on this server?
					online_users = []
					curr_users = [req.username for req in self.server.requests]
					print 'ison', users, curr_users
					for user in users:
						if user in curr_users:
							online_users.append(user)
						#end if
					#end if
					self.request.send(':%s 303 %s :%s\n' % (self.server.serverName, username, ' '.join(online_users)))
					noCmd = False
				#end if
				if 'PART' in cmds:
					state = 2
					self.partChannel(username, lines[cmds.index('PART')])
					noCmd = False
				#end if
				if 'PING' in cmds:
					# Reply to ever ping with a pong
					pingData = lines[cmds.index('PING')].split(' ')
					uniq = ''
					if len(pingData) > 1:
						uniq = pingData[1]
					#end if
					self.request.send('PONG %s\n' % uniq)
					noCmd = False
				#end if
				if 'NICK' in cmds:
					# Change NICK
					new_username = lines[cmds.index('NICK')][5:]
					# Ensure that the username isn't already taken.
					if new_username in self.server.irc_clients:
						self.request.send(':%s 433 * %s :Nickname is already in use.\n' % (self.server.serverName, new_username))
					else:
						self.request.send(':%s!~unknown@%s NICK :%s\n' % (username, self.host[0], new_username))
						username = new_username
						# Changed the username.
						self.username = username
					#end if
				#end if
				if noCmd:
					print 'doing nada in response ^_^'
				#end if
			#end if
		#loop
		print "%s:%d disconnected" % self.client_address
		# Remove ourself from the server list.
		a = [req == self for req in self.server.requests]
		if True in a:
			self.server.requests.pop(a.index(True))
		else:
			print 'We were not in self.server.requests!', a
		#end if
		a = [uname == self.username for uname in self.server.irc_clients]
		if True in a:
			self.server.irc_clients.pop(a.index(True))
		else:
			print 'We were not in self.server.irc_clients!', a
		#end if
		self.exiting = True
		for channel in self.server.channels:
			self.partChannel(self.username, 'PART ' + channel)
		#next channel
	#end def handle(self)
	
	def joinChannel(self, username, line):
		"""
		Join a channel.
		"""
		print 'joining with', line
		# Format: JOIN #neg9,&blah password
		data = line.split(' ')
		if len(data) < 2:
			print 'Warning: Invalid channel'
			return
		#end if
		channel = data[1]
		if channel in self.server.channels:
			# Add them to the channel.
			if username in self.server.channels[channel]:
				# They are trying to rejoin, rejected.
				print 'Warning: Already in channel'
				return
			#end if
			self.server.channels[channel].append(username)
		else:
			# Create a channel
			self.server.channels[channel] = [username]
		#end if
		msg_data = ':%s!~unknown@%s JOIN :%s\n' % (username, self.host[0], channel)
		self.request.send(msg_data)
		self.sendChannel(channel, msg_data)
		topic = 'Testing'
		topicMeta = 'Javantea!~r3wt@c-66-235-61-3.sea.wa.customer.broadstripe.net 1302504621'
		userlist = ' '.join(self.server.channels[channel])
		commonFields = [self.server.serverName, username, channel]
		self.request.send(':%s 332 %s %s :%s\n' % tuple(commonFields + [topic]))
		self.request.send(':%s 333 %s %s %s\n' % tuple(commonFields + [topicMeta]))
		self.request.send(':%s 353 %s = %s :%s\n' % tuple(commonFields + [userlist]))
		self.request.send(':%s 366 %s %s :End of /NAMES list.\n' % tuple(commonFields))

	#end def joinChannel(username, line)

	def partChannel(self, username, line):
		"""
		Leave a channel.
		"""
		print 'parting with', line
		# Format: PART #neg9
		data = line.split(' ')
		if len(data) < 2:
			print 'Invalid channel'
			return
		#end if
		channel = data[1]
		if channel not in self.server.channels:
			# Part from a channel that doesn't exist.
			print "Warning: Channel doesn't exist"
			return
		#end if
		if username not in self.server.channels[channel]:
			# They are trying to part a channel they aren't in, rejected.
			print 'Warning: Not in channel'
			return
		#end if
		# Remove them to the channel.
		self.server.channels[channel].pop(self.server.channels[channel].index(username))
		msg_data = ':%s!~unknown@%s PART %s\n' % (username, self.host[0], channel)
		# FIXME: This is not always set. =(
		if not self.exiting:
			self.request.send(msg_data)
		#end if
		self.sendChannel(channel, msg_data)
	#end def partChannel(username, line)

	def sendMsg(self, username, line):
		print 'sending message', line
		# Format: PRIVMSG #neg9 :aw yeah
		# Username and reverse lookup hostname
		user_data = ['unknown', self.host[0]]
		if self.user_data != None and len(self.user_data) > 0:
			user_data[0] = self.user_data[0]
		#end if
		cmd, channel, message = line.split(' ', 2)
		msg_data = ':%s!~%s@%s PRIVMSG %s %s\n' % (username, user_data[0], user_data[1], channel, message)
		self.sendChannel(channel, msg_data)
	#end def sendMsg(self, username, line)
	
	def sendChannel(self, channel, msg_data):
		if channel[0] != '#':
			return self.sendUser(channel, msg_data)
		#end if
		if channel not in self.server.channels:
			print "Error: Sending to a channel that doesn't exist"
			return
		#end if
		# Send to the list of users connected to our IRC Server.
		for req in self.server.requests:
			# Don't act upon ourself, that would be bad.
			if req == self:
				continue
			#end if
			# If this user is in the channel, send it to them.
			if req.username in self.server.channels[channel]:
				print 'sending to', req.username
				req.request.send(msg_data)
			#end if
		#next req
	#end def sendChannel(msg_data)
	
	def sendUser(self, user, msg_data):
		if user[0] == '#':
			# We don't send to channels.
			return
		#end if
		# Send to the list of users connected to our IRC Server.
		for req in self.server.requests:
			# Don't act upon ourself, that would be bad.
			if req == self:
				continue
			#end if
			# If this user is in the channel, send it to them.
			if req.username == user:
				req.request.send(msg_data)
			#end if
		#next req
	#end def sendChannel(msg_data)
#end class BikeIM_IRCHandler(SocketServer.BaseRequestHandler)

class Rss_IRCBot(ircbot.SingleServerIRCBot):
	def __init__(self, server_list, channel, nickname, realname):
		ircbot.SingleServerIRCBot.__init__(self, server_list, nickname, realname)
		self.running = None
		self.channel = channel
	#end def __init__(server_list, channel, nickname, realname)
	
	def shutdown(self):
		self.running = False
	#end def shutdown()
	
	def serve_forever(self):
		"""
		Directly from rssirc1.py by Javantea Mar  9  2008
		"""
		# if shutdown was called, we want to exit.
		if self.running == False: return
		# if shutdown has not been called, let it know that we're going.
		self.running = True

		self._connect()
		i = 0
		# FIXME: Get an exact solution.
		while i < 5 and self.running:
			# let it connect for 10 seconds or 5 commands.
			self.ircobj.process_once(2)
			i += 1
		#loop
		if self.running == False: return
		self.connection.join(self.channel)
		print 'Waiting until bot gets into channel:',
		i = 16
		# FIXME: Wait less than 600 seconds or 300 commands.
		while len(self.channels) == 0 and self.running:
			# process for 2 seconds, very nice.
			self.ircobj.process_once(2)
			print '.',
			if i % 40 == 0:
				print
				self.connection.join(self.channel)
			#end if
			if i > 300: break
			i += 1
		#loop
		if self.running == False: return
		print 'Got into the channel:', self.channels
		
		"""
		# If we did a link_start, then 
		should_sleep = (len(show_on_start) > 1)
		for (url, title) in show_on_start:
			self.connection.privmsg(self.channel, title + ' ' + url)
			if should_sleep: time.sleep(1)
		#next (url, title)
		if self.running == False: return
		"""
		
		# The main loop
		prev_time = 0
		while self.running:
			# process for 2 seconds, very nice.
			self.ircobj.process_once(2)
			#print 'bot processed.'
			
			# Every 60 seconds or so, we check rss.
			now = time.time()
			if now - prev_time > 60:
				prev_time = now
				#items = getNewRSS(rss_url)
				items = []
				#print 'got items', items
				should_sleep = (len(items) > 1)
				for (url, title) in items:
					self.connection.privmsg(self.channel, title + ' ' + url)
					if should_sleep: time.sleep(1)
				#next (url, title)
			#end if
		#loop
	#end def serve_forever()
#end class BikeIM_IRCBot(ircbot.SingleServerIRCBot)

class BikeIM_IRCBot(ircbot.SingleServerIRCBot):
	"""
	From ircbot/bitbot.py
	"""
	def __init__(self, channel, nickname, server, port, realName):
		ircbot.SingleServerIRCBot.__init__(self, [(server, port)], nickname, realName)
		self.channel = channel
		self.nickname = nickname
		self.queue = OutputManager(self.connection)
		self.badnicks = 0
		self.running = None
		self.mute = False
	#end def __init__(channel, nickname, server, port, realName)
	
	def start(self):
		"""
		Deal with the threads.
		"""
		# Sleep to allow the server to start hopefully.
		time.sleep(1.0)
		# The queue just allows us to send in a thread.
		self.queue.start()
		self._connect()
		self.serve_forever()
	#end def start()
	
	def shutdown(self):
		self.running = False
		# Very important because it won't exit without being disconnected.
		self.connection.disconnect()
	#end def shutdown()
	
	def serve_forever(self, timeout=0.2):
		"""Run a loop that exits on shutdown, processing data from connections.
		   This method repeatedly calls process_once.
		Arguments:
			timeout -- Parameter to pass to process_once.
		"""
		self.running = True
		while self.running:
			self.ircobj.process_once(timeout)
		#loop
	#end def serve_forever(timeout=0.2)
	
	def on_nicknameinuse(self, c, e):
		"""
		This becomes a loops quite easily. But after 4 tries we give up.
		"""
		if self.badnicks > 4:
			print 'Error: Giving up due to too many bad nicks'
			return
		#end if
		# Choose a different one each time.
		add_nick = '_123456789XYZW'
		self.badnicks += 1
		self.nickname = c.get_nickname() + add_nick[self.badnicks % len(add_nick)]
		c.nick(self.nickname)
	#end def on_nicknameinuse(c, e)

	def on_welcome(self, c, e):
		c.join(self.channel)
	#end def on_welcome(c, e)

	def on_privmsg(self, c, e):
		from_nick = nm_to_n(e.source())
		self.do_command(e, e.arguments()[0], from_nick)
	#end def on_privmsg(c, e)

	def on_pubmsg(self, c, e):
		from_nick = nm_to_n(e.source())
		#a = e.arguments()[0].split(":", 1)
		#if len(a) > 1 \
		#	and irc_lower(a[0]) == irc_lower(self.nickname):
		self.do_command(e, e.arguments()[0].strip(), from_nick)
		#end if
	#end def on_pubmsg(c, e)

	def say_public(self, text):
		"Print TEXT into public channel, for all to see."
		self.queue.send(text, self.channel)
	#end def say_public(text)

	def say_private(self, nick, text):
		"Send private message of TEXT to NICK."
		self.queue.send(text,nick)
	#end say_private(nick, text)

	def reply(self, text, to_private=None):
		"Send TEXT to either public channel or TO_PRIVATE nick (if defined)."

		if to_private is not None:
			self.say_private(to_private, text)
		else:
			# Mute turns off all public communication.
			if self.mute:
				return
			#end if
			self.say_public(text)
		#end if
	#end def reply(text, to_private=None)

	def usage(self):
		"""
		Print Usage.
		"""
		return "messages | message $i | delete $i | m | m $i | mark $i  | md $i | mute | unmute"
	#end def usage()

	def messages(self):
		"Return a list of messages."
		global messages
		return ('messages: %d.' % len(messages)) + ' ' + ', '.join([('%d: %s' % (m.id, m.subject)) for m in messages])
	#end def messages()

	def message(self, i):
		"Return a single message."
		global messages
		if i.strip() == '': return self.messages()
		try:
			j = int(i.strip())
		except ValueError:
			return 'Invalid message'
		#end try
		for message in messages:
			if message.id == j:
				return '%d *%s* %s' % (message.id, message.subject, message.body)
			#end if
		#next message
		return 'message %d not found' % j
	#end def message(i)

	def url(self, i):
		"Return a message's url."
		global messages
		server_ip = 'localhost'
		try:
			j = int(i.strip())
		except ValueError:
			return 'http://%s:1234/messages/' % server_ip
		#end try
		for message in messages:
			if message.id == j:
				return 'http://%s:1234/messages/%d' % (server_ip, message.id)
			#end if
		#next message
		return 'http://%s:1234/messages/' % server_ip
	#end def url(i)

	def mark(self, i, marked_by):
		"Mark a message owned by a user."
		global messages
		try:
			j = int(i.strip())
		except ValueError:
			return 'Cannot mark a non-integer.'
		#end try
		for message in messages:
			if message.id == j:
				message.owner = marked_by
				saveMessages(messages)
				return '%d marked by %s' % (message.id, marked_by)
			#end if
		#next message
		return "Message %d doesn't exist, try again?" % j
	#end def mark(i, marked_by)

	def delete(self, i):
		"Delete a message."
		global messages
		try:
			j = int(i.strip())
		except ValueError:
			return 'Cannot delete a non-integer.'
		#end try
		i = 0
		for message in messages:
			if message.id == j:
				messages.pop(i)
				saveMessages(messages)
				return '%d Deleted' % (message.id)
			#end if
			i += 1
		#next message
		return "Message %d doesn't exist, try again?" % j
	#end def delete(i)

	def do_command(self, e, cmd, from_private):
		"""This is the function called whenever someone sends a public or
		private message addressed to the bot. (e.g. "bot: blah").  Parse
		the CMD, execute it, then reply either to public channel or via
		/msg, based on how the command was received.  E is the original
		event, and FROM_PRIVATE is the nick that sent the message."""

		if e.eventtype() == "pubmsg":
			# self.reply() sees 'from_private = None' and sends to public channel.
			target = None
		else:
			# assume that from_private comes from a 'privmsg' event.
			target = from_private.strip()
		#end if

		if cmd.startswith('help'):
			self.reply(self.usage(), target)
		elif cmd.startswith('messages'):
			self.reply(self.messages(), target)
		elif cmd.startswith('message'):
			self.reply(self.message(cmd[7:]), target)
		elif cmd == 'm':
			self.reply(self.messages(), target)
		elif cmd.startswith('m '):
			self.reply(self.message(cmd[1:]), target)
		elif cmd.startswith('md'):
			self.reply(self.delete(cmd[2:]), target)
		elif cmd.startswith('delete'):
			self.reply(self.delete(cmd[6:]), target)
		elif cmd.startswith('url'):
			self.reply(self.url(cmd[3:]), target)
		elif cmd.startswith('mark'):
			self.reply(self.mark(cmd[4:], from_private), target)
		elif cmd.startswith('mute'):
			self.mute = True
		elif cmd.startswith('unmute'):
			self.mute = False
		#end if
	#end def do_command(e, cmd, from_private)
#end class BikeIM_IRCBot(ircbot.SingleServerIRCBot)

class OutputManager(threading.Thread):
	"""
	From ircbot/botcommon.py
	"""
	def __init__(self, connection, delay=.5):
		threading.Thread.__init__(self)
		self.setDaemon(1)
		self.connection = connection
		self.delay = delay
		self.event = threading.Event()
		self.queue = []
	#end def __init__(self, connection, delay=.5)
	
	def run(self):
		while 1:
			self.event.wait()
			while self.queue:
				msg,target = self.queue.pop(0)
				self.connection.privmsg(target, msg)
				time.sleep(self.delay)
			self.event.clear()
	#end def run()

	def send(self, msg, target):
		self.queue.append((msg.strip(),target))
		self.event.set()
	#end def send(msg, target)
#end class OutputManager(threading.Thread)


# Get messages from a file.
currentMaxMsgId = 1
messages = loadMessages()

def main():
	global ircb
	address = ''
	http_port = 1234
	irc_port = 6667
	httpd = BikeIM_HTTPServer((address, http_port))
	ircd  = BikeIM_IRCServer((address, irc_port))
	ircb  = BikeIM_IRCBot('#sakuracon', 'sakreghelp', 'localhost', 6667, 'Sakura Reg Help Bot')
	print 'HTTP server', http_port
	httpd_thread = threading.Thread(target=httpd.serve_forever)
	httpd_thread.start()
	print 'IRC server', irc_port
	ircd_thread = threading.Thread(target=ircd.serve_forever)
	ircd_thread.start()
	print 'IRC bot'
	ircb_thread = threading.Thread(target=ircb.start)
	ircb_thread.start()
	try:
		while True:
			time.sleep(1000.0)
		#loop
	except KeyboardInterrupt:
		print 'User Interrupted.'
	#end try
	print 'httpd please go down'
	httpd.shutdown()
	httpd_thread.join()
	print 'ircd please go down'
	ircd.shutdown()
	ircd_thread.join()
	print 'irc bot please go down'
	ircb.shutdown()
	ircb_thread.join()
	print 'all down'
#end def main()

if __name__ == '__main__':
	main()
#end if
